/*
 * USB-MIDI class driver for GR-SAKURA
 * Copyright 2012-2014 Yuuichi Akagawa
 *
 *******************************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 *******************************************************************************
 */
#include "usbh_midi.h"

const byte USBH_MIDI::epDataInIndex     = 1;
const byte USBH_MIDI::epDataOutIndex    = 2;
const byte USBH_MIDI::epDataInIndexVSP  = 3;
const byte USBH_MIDI::epDataOutIndexVSP = 4;
const int  USBH_MIDI::usbTimeout = 0;    //BulkTransfer timeout(ms)

/* Constructor */
USBH_MIDI::USBH_MIDI(bool enableLED)
{
    isLedEnable = enableLED;
    isPollEnable = false;
    isMidiFound = false;
    readPtr = 0;
    bConfNum    = 0; //Configuration descriptor number
    Init();
}

/* Connection initialization of an MIDI Device */
void USBH_MIDI::Init()
{
    if( isLedEnable ){
        pinMode(PIN_LED0,OUTPUT);
    }

    if(tkusbh_init() != TKUSBH_OK)
    {
        //USB Host is not available
        pinMode(PIN_LED2,OUTPUT);
        pinMode(PIN_LED3,OUTPUT);
        while(1){
            digitalWrite(PIN_LED2, HIGH);
            digitalWrite(PIN_LED3, LOW);
            delay(100);
            digitalWrite(PIN_LED2, LOW);
            digitalWrite(PIN_LED3, HIGH);
        }
    }
    for(int i=0; i<MIDI_MAX_ENDPOINTS; i++){
        epInfo[i].epAddr = 0;
    }
}

/* Performs a cleanup after disconnect */
void USBH_MIDI::Release()
{
    tkusbh_disconnect(TIMEOUT_INFINITE);
    isPollEnable = false;
    bConfNum     = 0;
    readPtr = 0;
    for(int i=1; i<MIDI_MAX_ENDPOINTS; i++){
        epInfo[i].epAddr = 0;
    }

    if( isLedEnable ){
        digitalWrite(PIN_LED0, LOW);
    }
    return;
}

/* Check and Connect to MIDI device */
bool USBH_MIDI::isConnected()
{
    TKUSBH_RESULT rcode;
    //Check the disconnect state
    if(isPollEnable && tkusbh_is_connected() == TKUSBH_DISCONNECT)
    {
        Release();        
        return false;
    }

    //Already connected
    if(isPollEnable)
        return true;

    //Start Enumeration
     if(tkusbh_connect(500) != TKUSBH_OK)
     {
         return false;
     }

    //Enumeration done. parse configuration descriptor and set endpoint
    byte num_of_conf = USBHostInfo.DeviceDesc.bNumConfigurations;
     for (byte i=0; i<num_of_conf; i++) {
         if( parseConfigDescr(i) )    //Midi found?
             break;
     }
 
    if( bConfNum == 0 ){  //Device not found.
         Release();
         return false;
     }
 
     if( !isMidiFound ){ //MIDI Device not found. Try first Bulk transfer device
         epInfo[epDataInIndex].epAddr      = epInfo[epDataInIndexVSP].epAddr;
         epInfo[epDataInIndex].maxPktSize  = epInfo[epDataInIndexVSP].maxPktSize;
         epInfo[epDataOutIndex].epAddr     = epInfo[epDataOutIndexVSP].epAddr;
         epInfo[epDataOutIndex].maxPktSize = epInfo[epDataOutIndexVSP].maxPktSize;
     }
 
     // Set Configuration Value
     rcode = tkusbh_set_configuration(bConfNum);
     if (rcode == TKUSBH_ERROR) {
         Release();
        return false;
     }
     
     //Connect done.
     isPollEnable = true;
    if( isLedEnable ){
        digitalWrite(PIN_LED0, HIGH);
    }

    delay(300);
    return true;
}

/* get and parse config descriptor */
bool USBH_MIDI::parseConfigDescr( byte conf )
{
    byte buf[ DESC_BUFF_SIZE ];
    byte *buf_ptr = buf;
    byte descr_length;
    byte descr_type;
    unsigned short total_length;
    EndpointDesc_t *epDesc;
    bool isMidi = false;
    byte ifNum;
    byte confNum;

    //check descritpor size
    total_length = USBHostInfo.ConfigDesc.wTotalLength;
    if( total_length > DESC_BUFF_SIZE ) {    //check if total length is larger than buffer
        total_length = DESC_BUFF_SIZE;
    }

    // get configuration descriptor (all)
    if( tkusbh_get_descriptor(CONFIG_DESCRIPTOR_TYPE, conf, buf, total_length) < 0 ){
        return(false);
    }  

    //parsing descriptors
    while( buf_ptr < buf + total_length ) {  
        descr_length = buf_ptr[0];
        descr_type   = buf_ptr[1];
        switch( descr_type ) {
            case CONFIG_DESCRIPTOR_TYPE :
                confNum = buf_ptr[5];
                break;
            case  INTERFACE_DESCRIPTOR_TYPE :
//                ifNum = buf_ptr[2];
                if( buf_ptr[5] == USB_CLASS_AUDIO && buf_ptr[6] == USB_SUBCLASS_MIDISTREAMING ) {  //p[5]; bInterfaceClass = 1(Audio), p[6]; bInterfaceSubClass = 3(MIDI Streaming)
                    isMidiFound = true; //MIDI device found.
                    isMidi      = true;
                    bConfNum    = confNum;
                }else{
                    isMidi = false;
                }
                break;
            case ENDPOINT_DESCRIPTOR_TYPE :
                epDesc = (EndpointDesc_t *)buf_ptr;
                if ((epDesc->bmAttributes & 0x02) == 2) {//bulk
                    byte index;
                    if( isMidi ){
                        index = ((epDesc->bEndpointAddress & 0x80) == 0x80) ? epDataInIndex : epDataOutIndex;
                    }else{
                        index = ((epDesc->bEndpointAddress & 0x80) == 0x80) ? epDataInIndexVSP : epDataOutIndexVSP;
                        if( isMidiFound == false ){
                            bConfNum = confNum;
                        }
                    }
                    if( epInfo[index].epAddr == 0 ){
                        epInfo[index].epAddr     = (epDesc->bEndpointAddress & 0x0F);
                        epInfo[index].maxPktSize = epDesc->wMaxPacketSize;
                        epInfo[index].bInterval  = epDesc->bInterval;
                    }
                }
                break;
            default:
                break;
        }//switch( descr_type  
        buf_ptr += descr_length;    //advance buffer pointer
    }//while( buf_ptr <=...
    return(isMidiFound);
}

/* Receive USB-MIDI Event Packets */
int USBH_MIDI::rawRead(byte *dataptr)
{
    return tkusbh_bulk_read(epInfo[epDataInIndex].epAddr, dataptr, (int)epInfo[epDataInIndex].maxPktSize, usbTimeout);
}

/* Receive data from MIDI device */
int  USBH_MIDI::read(byte *dataptr)
{
    int  rcode;

    if( isPollEnable == false ) return false;

    //Checking unprocessed message in buffer.
    if( readPtr != 0 && readPtr < MIDI_EVENT_PACKET_SIZE ){
        if(recvBuf[readPtr] == 0 && recvBuf[readPtr+1] == 0) {
            //no unprocessed message left in the buffer.
        }else{
            goto RecvData_return_from_buffer;
        }
    }

    readPtr = 0;
    rcode = rawRead(recvBuf);
    if( rcode < 0 ) {
        return false;
    }
  
    //if all data is zero, no valid data received.
    if( recvBuf[0] == 0 && recvBuf[1] == 0 && recvBuf[2] == 0 && recvBuf[3] == 0 ) {
        return false;
    }
RecvData_return_from_buffer:
    //Extract MIDI messages
    readPtr++;
    dataptr[0] = recvBuf[readPtr];    //MIDI_0
    readPtr++;
    dataptr[1] = recvBuf[readPtr];    //MIDI_1
    readPtr++;
    dataptr[2] = recvBuf[readPtr];    //MIDI_2
    readPtr++;
    return true;
}

/* Send data to MIDI device */
/* nCable is default = 0 */
int USBH_MIDI::write(byte *dataptr, byte nCable)
{
    byte buf[4];
    byte msg;

    msg = dataptr[0];
    // SysEx long message ?
    if( msg == 0xf0 )
    {
        return writeSysEx(dataptr, countSysExDataSize(dataptr), nCable);
    }

    buf[0] = (nCable << 4) | (msg >> 4);
    if( msg < 0xf0 ) msg = msg & 0xf0;

    //Building USB-MIDI Event Packets
    buf[1] = dataptr[0];
    buf[2] = dataptr[1];
    buf[3] = dataptr[2];
    switch(lookupMsgSize(msg)) {
        //3 bytes message
        case 3 :
            if(msg == 0xf2) {//system common message(SPP)
                buf[0] = (nCable << 4) | 3;
            }
            break;

        //2 bytes message
        case 2 :
            if(msg == 0xf1 || msg == 0xf3) {//system common message(MTC/SongSelect)
                buf[0] = (nCable << 4) | 2;
            }
            buf[3] = 0;
            break;

        //1 bytes message
        case 1 :
        default :
            buf[2] = 0;
            buf[3] = 0;
            break;
    }
    return tkusbh_bulk_write(epInfo[epDataOutIndex].epAddr, buf, 4, usbTimeout);
}

/* look up a MIDI message size from spec */
/*Return                                 */
/*  0 : undefined message                */
/*  0<: Vaild message size(1-3)          */
int USBH_MIDI::lookupMsgSize(byte midiMsg)
{
    int msgSize = 0;

    if( midiMsg < 0xf0 ) midiMsg &= 0xf0;
    switch(midiMsg) {
        //3 bytes messages
        case 0xf2 : //system common message(SPP)
        case 0x80 : //Note off
        case 0x90 : //Note on
        case 0xa0 : //Poly KeyPress
        case 0xb0 : //Control Change
        case 0xe0 : //PitchBend Change
            msgSize = 3;
            break;

        //2 bytes messages
        case 0xf1 : //system common message(MTC)
        case 0xf3 : //system common message(SongSelect)
        case 0xc0 : //Program Change
        case 0xd0 : //Channel Pressure
            msgSize = 2;
            break;

        //1 bytes messages
        case 0xf8 : //system realtime message
        case 0xf9 : //system realtime message
        case 0xfa : //system realtime message
        case 0xfb : //system realtime message
        case 0xfc : //system realtime message
        case 0xfe : //system realtime message
        case 0xff : //system realtime message
            msgSize = 1;
            break;

        //undefine messages
        default :
            break;
    }
    return msgSize;
}

/* Send SysEx message to MIDI device */
/* nCable is default = 0 */
int USBH_MIDI::writeSysEx(byte *dataptr, word datasize, byte nCable)
{
    byte buf[4];
    int rc;
    word n = datasize;

    //Byte 0
    buf[0] = (nCable << 4) | 0x4;         //x4 SysEx starts or continues

    while(n > 0) {
        switch ( n ) {
        case 1 :
            buf[0] = (nCable << 4) | 0x5;   //x5 SysEx ends with following single byte.
            buf[1] = *(dataptr++);
            buf[2] = 0x00;
            buf[3] = 0x00;
            n = n - 1;
            break;
        case 2 :
            buf[0] = (nCable << 4) | 0x6;   //x6 SysEx ends with following two bytes.
            buf[1] = *(dataptr++);
            buf[2] = *(dataptr++);
            buf[3] = 0x00;
            n = n - 2;
            break;
        case 3 :
            buf[0] = (nCable << 4) | 0x7;   //x7 SysEx ends with following three bytes.
        default :
            buf[1] = *(dataptr++);
            buf[2] = *(dataptr++);
            buf[3] = *(dataptr++);
            n = n - 3;
            break;
        }
        rc = tkusbh_bulk_write(epInfo[epDataOutIndex].epAddr, buf, 4, usbTimeout);
        if(rc == -1) break;
    }
    return(rc);
}

/* SysEx data size counter */
int USBH_MIDI::countSysExDataSize(byte *dataptr)
{
    int c = 1;

    if( *dataptr != 0xf0 ){ //not SysEx
      return 0;
    }

    //Search terminator(0xf7)
    while(*dataptr != 0xf7)
    {
        dataptr++;
        c++;
    
        //Limiter (upto 256 bytes)
        if(c > 256){
            c = 0;
            break;
        }
    }
    return c;
}
